#include <urbi/uobject.hh>
#include <urbi/uconversion.hh>

#include <cv.h>
#include <highgui.h>

#include <cstdlib>
#include <iostream>
#include <sstream>
#include <string>
#include <vector>

#include "stitchPan.hpp"
#include "calibration.hpp"
#include "mask.hpp"

using namespace std;
using namespace cv;
using namespace urbi;

class ArPan : public UObject
{
public:
	ArPan(const string&);
	~ArPan();
	
	int init(string source);

private:
    virtual int update(); // Called periodically.
    void showFrame();
    void getFrame();
    void saveFrame(string suffix);
    void updateFrameNo();
    void makePanorama();
    int getKey();
    
    UVar* imagesource;     // Pointer to UVar that contains our desired image.
    UBinary src; // Binary storage for the image.
    IplImage* cvimg;
    Mat image;
    
    vector<Mat> storedImages;
    
    int imNum;
};

// Constructor: bind the init function, and leave everything else nulled.
ArPan::ArPan(const string& s)
    : UObject(s), imagesource(NULL), cvimg(NULL), imNum(0)
{
    UBindFunctions(ArPan, init, showFrame, saveFrame,
                   updateFrameNo, makePanorama, getKey);
}

// Destructor: delete the image header.
ArPan::~ArPan()
{
    cvReleaseImageHeader(&cvimg);
}

// Update function handles the OpenCV window to prevent it from freezing.
int ArPan::update()
{
    waitKey(1);
    return 0;
}

// Initialize the drone, lock on to the camera feed, and set update to run every
// so often.
int ArPan::init(string source)
{
    initializeCalibration();
    imagesource = new UVar(source);
    
    // Every time the image changes, get the new frame.
    UNotifyChange(*imagesource, &ArPan::getFrame);
    
    // Update every 10ms (to keep window from lagging out)
    USetUpdate(10);

    cvNamedWindow("Video Feed", CV_WINDOW_AUTOSIZE);
    cvMoveWindow("Video Feed", 0, 10);
    
    return 0;
}

void ArPan::showFrame()
{   
    if (cvimg)
        imshow("Video Feed", image);
}

void ArPan::getFrame()
{
    UVar& imRef = *imagesource;
    src = imRef;

    // Allocate room for the incoming image.
    if (!cvimg)
        cvimg = cvCreateImageHeader(cvSize(src.image.width,
                                           src.image.height),
                                    IPL_DEPTH_8U, 3);
    
    // Save the image (well, move around a pointer to the image)    
    cvimg->imageData = (char*)src.image.data;
    cvCvtColor(cvimg, cvimg, CV_RGB2BGR);
    
    image = cvimg;
    
    showFrame();
}

void ArPan::saveFrame(string suffix)
{
    stringstream out;
    out << imNum;
    string fileNum = out.str();
    
    // Save the image to file.
    while (fileNum.length() < 3)
        fileNum = "0" + fileNum;
        
    string filename = "images/" + fileNum + suffix + ".jpg";
    imwrite(filename.c_str(), image);
    
    // Push an undistorted image into the set that we're merging.
    Mat undistorted;
    undistort(addMask(image), undistorted);
    storedImages.push_back(undistorted);
}

void ArPan::makePanorama()
{
    cvNamedWindow("combined", CV_WINDOW_AUTOSIZE);
    cvMoveWindow("combined", 0, 400);
    Mat combined = stitchPanI(storedImages);
    imwrite("panorama.png", combined);
}

void ArPan::updateFrameNo()
{
    imNum++;
}

int ArPan::getKey()
{
    return waitKey(10);
}

UStart(ArPan);