#include "corner.hpp"
#include "helpers.hpp"
#include "homography.hpp"
#include "mask.hpp"
#include "match.hpp"

using namespace cv;
using namespace std;

Mat applyHomography(Mat& H, Mat& quad)
{
    Mat newCorners = H * quad;
    for(int i = 0; i < quad.cols; ++i) {
        newCorners.at<double>(0, i) /= newCorners.at<double>(2, i);
        newCorners.at<double>(1, i) /= newCorners.at<double>(2, i);
        newCorners.at<double>(2, i) = 1;
    }
    return newCorners;
}

Point2f applyHomography(Mat& H, Point2f pt)
{
    Mat tempPt(3, 1, CV_64F);
    tempPt.at<double>(0, 0) = pt.x;
    tempPt.at<double>(1, 0) = pt.y;
    tempPt.at<double>(2, 0) = 1;
    
    Mat outMat = applyHomography(H, tempPt);
    return Point2f(outMat.at<double>(0, 0), outMat.at<double>(1, 0));
}

Mat getDistortedCorners(Size size, Mat& H)
{
    Mat corner = Mat(3, 4, CV_64F);
    Point2f origQuad[4] = {Point2f(0, 0),
                           Point2f(size.width, 0),
                           Point2f(size.width, size.height),
                           Point2f(0, size.height)};
    for(int i = 0; i < 4; ++i) {
        corner.at<double>(0, i) = origQuad[i].x;
        corner.at<double>(1, i) = origQuad[i].y;
        corner.at<double>(2, i) = 1;
    }
    
    Mat newCorners = applyHomography(H, corner);
    return newCorners;
}

double getDistortedAreaFrac(Size size, Mat& H)
{
    Mat newCorners = getDistortedCorners(size, H);
    
    // Magic formula from:
    // http://softsurfer.com/Archive/algorithm_0101/algorithm_0101.htm
    double area = (newCorners.at<double>(0, 2) - newCorners.at<double>(0, 0)) *
                  (newCorners.at<double>(1, 3) - newCorners.at<double>(1, 1)) -
                  (newCorners.at<double>(0, 3) - newCorners.at<double>(0, 1)) *
                  (newCorners.at<double>(1, 2) - newCorners.at<double>(1, 0));
    
    double minX, minY, maxX, maxY;
    minMaxLoc(newCorners.row(0), &minX, &maxX);
    minMaxLoc(newCorners.row(1), &minY, &maxY);

    double bbarea = (maxX - minX) * (maxY - minY);

    return abs(0.25 * area) / bbarea;
}

Mat combine(Mat& H, Mat& im1, Mat& im2, Mat& G, Point2f& offset)
{   
    if (im1.type() == CV_8UC3)
    {
        cerr << "Warning: Unmasked image coming into combine" << endl;
        im1 = addMask(im1);
    }
    if (im2.type() == CV_8UC3)
    {
        cerr << "Warning: Unmasked image coming into combine" << endl;
        im2 = addMask(im2);
    }
    
    // Save the corners!
    Point2f origQuad[4] = {Point2f(0, 0),
                           Point2f(im1.cols, 0),
                           Point2f(im1.cols, im1.rows),
                           Point2f(0, im1.rows)};

    // Project the corners using the homography.  Then, figure out the bounding
    // box of the projection.
    Mat newCorners = getDistortedCorners(im1.size(), H);

    double minX, minY, maxX, maxY;
    minMaxLoc(newCorners.row(0), &minX, &maxX);
    minMaxLoc(newCorners.row(1), &minY, &maxY);
    minX = min(0.0, minX); maxX = max(double(im2.cols), maxX);
    minY = min(0.0, minY); maxY = max(double(im2.rows), maxY);
    
    Size newSize = Size(maxX-minX, maxY-minY);
	offset = Point2f(-minX, -minY);
    
    // If the reprojected image is too large, then shrink it down in size.
    Size maxSize = Point2i(im1.size() + im2.size());
    if (newSize.width > maxSize.width)
    {
        double difference = newSize.width - maxSize.width;
        offset.x = max(0.0, offset.x - difference);
        newSize.width = maxSize.width;
    }
    if (newSize.height > maxSize.height)    
    {
        double difference = newSize.height - maxSize.height;
        offset.y = max(0.0, offset.y - difference);
        newSize.height = maxSize.height;
    }

    // Figure out a new homography that will map into the boundaries of a larger
    // image cleanly, without pixels popping up at negative coordinates.
    Point2f newQuad[4] = {Point2f(newCorners.at<double>(0, 0),
                                  newCorners.at<double>(1, 0)) + offset,
                          Point2f(newCorners.at<double>(0, 1),
                                  newCorners.at<double>(1, 1)) + offset,
                          Point2f(newCorners.at<double>(0, 2),
                                  newCorners.at<double>(1, 2)) + offset,
                          Point2f(newCorners.at<double>(0, 3),
                                  newCorners.at<double>(1, 3)) + offset};
    G = getPerspectiveTransform(origQuad, newQuad);
    
    // Do the actual warping.
    Mat warped1, warped2;
    warpPerspective(im1, warped1, G, newSize);
    warped2 = expand(im2, newSize, offset);

    Mat out = combineMasked(warped1, warped2);

    return out;
}

Mat findBestHomography(vector<Match>& matches)
{
    // What is an inlier?
    double THRESHOLD = 4;
    
    // Copy points into a matrix.
    Mat srcPoints(Size(1, matches.size()), CV_32FC2),
        dstPoints(Size(1, matches.size()), CV_32FC2);
        
    for(size_t i = 0; i < matches.size(); ++i)
    {
        srcPoints.at<Point2f>(0, i) = matches[i].loc1;
        dstPoints.at<Point2f>(0, i) = matches[i].loc2;
    }
    
    // Use OpenCV's built-in RANSAC to find the homography.
    Mat H = findHomography(srcPoints, dstPoints, CV_RANSAC, THRESHOLD);
    return H;
}

Mat findBestHomography(Mat& im1, Mat& im2)
{
    vector<Point2f> corners1 = findCorners(im1);    
    vector<Point2f> corners2 = findCorners(im2);    

    vector<Feature> features1 = getFeatures(im1, corners1);
    vector<Feature> features2 = getFeatures(im2, corners2);
    
    std::vector<Match> matches = getMatches(features1, features2);
    
    return findBestHomography(matches);
}