#include "cylinder.hpp"
#include "homography.hpp"
#include "match.hpp"
#include "mask.hpp"
#include "rotation.hpp"

#include <map>

using namespace cv;
using namespace std;

Mat rotateImage(const Mat& source, double angle)
{
    Mat rotationMatrix = getRotationMatrix2D(Point2f(source.cols/2.0f,
                                                     source.rows/2.0f),
                                             angle, 1.0);
    Mat rotated;
    warpAffine(source, rotated, rotationMatrix, source.size());
    
    return tightenMask(rotated);
}

Point2f rotatePoint(Point2f pt, Point2f center, double rot)
{
    Mat rotationMatrix = getRotationMatrix2D(center, rot, 1.0);
    
    Mat tempPt(3, 1, CV_64F);
    tempPt.at<double>(0, 0) = pt.x;
    tempPt.at<double>(1, 0) = pt.y;
    tempPt.at<double>(2, 0) = 1;
    
    Mat rotPt = rotationMatrix * tempPt;
    return Point2f(rotPt.at<double>(0, 0), rotPt.at<double>(1, 0));
}


// We assume that the second set of locations in matches is already distorted,
// then solve for the best rotation.
double findBestRotation(vector<Match> matches, Size size, Mat& bestH)
{    
    std::vector<Match> rotatedMatches = matches;  
    
    // Miscellaneous state saving variables
    double bestRot = 0;
    double minDistortion = DBL_MAX;
    
    for (double rot = -4;
                rot < 4;
                rot += 0.1)
    {
        Point2f center(size.width/2.0f, size.height/2.0f);
        for(size_t v = 0; v < matches.size(); ++v)
        {
            // Rotate the point around by a certain amount.
            Point2f rotPt = rotatePoint(matches[v].loc1, center, rot);
            
            // Then convert it into cylindrical coordinates.
            rotatedMatches[v].loc1 = ToCylinder(rotPt, size, size);
        }
        
        // Find the new homography
        Mat H = findBestHomography(rotatedMatches);
            
        double distortion = 1 - getDistortedAreaFrac(size, H);
        if (abs(distortion) < abs(minDistortion))
        {
            bestRot = rot;
            minDistortion = distortion;
            H.copyTo(bestH);
        }
        
    }
    return bestRot;
}

Mat findSingleHomography(vector<Match> matches, Size size, double rot)
{
    std::vector<Match> rotatedMatches = matches; 
    Point2f center(size.width/2.0f, size.height/2.0f);
    
    for(size_t v = 0; v < matches.size(); ++v)
    {
        // Rotate the point around by a certain amount.
        Point2f rotPt = rotatePoint(matches[v].loc1, center, rot);
        
        // Then convert it into cylindrical coordinates.
        rotatedMatches[v].loc1 = ToCylinder(rotPt, size, size);
    }
    
    // Find the new homography
    return findBestHomography(rotatedMatches);
}

// This function is nice if you're running a bunch of rotation matching operations.
// If you're only doing it once, then it's not much help.
double findBestRotationCached(size_t im1ID, size_t im2ID, double im1rot,
                              vector<Match> matches, Size size, Mat& bestH)
{
/* with caching:
real	2m30.338s
user	2m8.196s
sys	0m10.341s
*/
    
    static vector<RotID> cache;
    static int queries = 0;
    static int hits = 0;
    
    queries++;
    RotID rid(im1ID, im2ID, im1rot);
    
    for(size_t r = 0; r < cache.size(); ++r)
        if (cache[r] == rid) {
            cout << "Cache hit" << endl;
            hits++;
            double rot = cache[r].im2rot;
            bestH = findSingleHomography(matches, size, rot);
//            cout << "Hit rate: " << double(hits) / double(queries) << endl;
            return rot;
        }

//    cout << "Cache miss!" << endl;
    rid.im2rot = findBestRotation(matches, size, bestH);
    
    cache.push_back(rid);
//    cout << "Hit rate: " << double(hits) / double(queries) << endl;
    return rid.im2rot;
}

RotID::RotID(size_t im1, size_t im2, double rot)
    : im1num(im1), im2num(im2), im1rot(rot), im2rot(0)
{
    // Nothing to do here.
}

bool operator==(const RotID& r1, const RotID& r2)
{
    return (r1.im1num == r2.im1num) && (r1.im2num == r2.im2num)
                                    && (r1.im1rot == r2.im1rot);
}