#include "stitchPan.hpp"

using namespace cv;
using namespace std;

Mat stitchPan(vector<Mat> images)
{
    vector<Feature> featuresCurr = getFeatures(images[0]),
                    featuresNext;
   
    // Save the original image size
    Mat combined = cylinderize(images[0]);

    // Pre-distort the original image cylindrically.
    for (size_t f = 0; f < featuresCurr.size(); ++f)
        featuresCurr[f].loc = ToCylinder(featuresCurr[f].loc,
                                         images[0].size(),
                                         images[0].size());
    
    imshow("out", maskedTo8U(combined));
    cout << "Processing images:" << endl;
    
    for (size_t i = 0; i + 1 < images.size(); ++i)
    {
            cout << "   " << i + 1 << "/"
                 << images.size() - 1 << endl;
            
            // Find features on the next image.
            featuresNext = getFeatures(images[i+1]);
            
            // Calculate matches.
            vector<Match> matches = getMatches(featuresNext, featuresCurr);

            // Find image rotation and ideal homography.
            Mat H, G;
            double rot = findBestRotation(matches, images[i+1].size(), H);
            
            // Rotate images and combine.
            Mat imNext = cylinderize(rotateImage(images[i+1], rot));

            Point2f offset;
            combined = combine(H, imNext, combined, G, offset);
            waitKey(1);
            imshow("out", maskedTo8U(combined));

            // Distort the next image's extracted features.
            Point2f center(images[i + 1].cols / 2.0, images[i + 1].rows / 2.0);
            for (size_t f = 0; f < featuresNext.size(); ++f)
            {   
                featuresNext[f].loc = rotatePoint(featuresNext[f].loc,
                                                  center, rot);
                featuresNext[f].loc = ToCylinder(featuresNext[f].loc,
                                                 images[i + 1].size(),
                                                 images[i + 1].size());
                featuresNext[f].loc = applyHomography(G, featuresNext[f].loc);
            }
            
            // Shift things along in the series.
            featuresCurr = featuresNext;
    }
    
    combined = maskedTo8U(combined);
    return tightCrop(combined);
}

Mat stitchPanI(vector<Mat> images)
{
    return stitchPanI(images, images.size() / 2);
}

Mat stitchPanI(vector<Mat> images, size_t imageNum)
{
    // Sanity checking
    if (imageNum >= images.size())
        return Mat();

    vector<Feature> featuresCurr = getFeatures(images[imageNum]),
                    featuresNext, featuresOrig;

    // Pre-distort the original image cylindrically.
    Mat combined = cylinderize(images[imageNum]);
    for (size_t f = 0; f < featuresCurr.size(); ++f)
        featuresCurr[f].loc = ToCylinder(featuresCurr[f].loc,
                                         images[imageNum].size(),
                                         images[imageNum].size());
    
    // Save these original features - we're going to need them later.
    featuresOrig = featuresCurr;
    
    Point2i totalOffset(0, 0);
    
    double prevRot = 0;
    
    // Build out the panorama to the left
    for (int i = imageNum; i > 0; --i)
    {       
            // Find features on the next image.
            featuresNext = getFeatures(images[i-1]);
                        
            // Calculate matches.
            vector<Match> matches = getMatches(featuresNext, featuresCurr);

            // Find image rotation and ideal homography.
            Mat H, G;
            double rot = findBestRotationCached(i, i-1, prevRot,
                                                matches, images[i-1].size(), H);
            
            // Rotate images and combine.
            Mat imNext = cylinderize(rotateImage(images[i-1], rot));

            Point2f offset;
            combined = combine(H, imNext, combined, G, offset);
            totalOffset += Point2i(int(offset.x), int(offset.y));
            
            // Show the image
            waitKey(1);
            imshow("combined", maskedTo8U(combined));

            // Distort the next image's extracted features.
            Point2f center(images[i - 1].cols / 2.0, images[i - 1].rows / 2.0);
            for (size_t f = 0; f < featuresNext.size(); ++f)
            {   
                featuresNext[f].loc = rotatePoint(featuresNext[f].loc,
                                                  center, rot);
                featuresNext[f].loc = ToCylinder(featuresNext[f].loc,
                                                 images[i - 1].size(),
                                                 images[i - 1].size());
                featuresNext[f].loc = applyHomography(G, featuresNext[f].loc);
            }
            
            // Shift things along in the series.
            featuresCurr = featuresNext;
            
            prevRot = rot;
    }
    
    // Now adjust the original set of features and stitch the other set of
    // images together.
    for (size_t f = 0; f < featuresOrig.size(); ++f)
        featuresOrig[f].loc += Point2f(totalOffset.x, totalOffset.y);
    featuresCurr = featuresOrig;
    prevRot = 0;
    
    // Build out the panorama to the right
    for (size_t i = imageNum; i + 1 < images.size(); ++i)
    {
            // Find features on the next image.
            featuresNext = getFeatures(images[i+1]);
            
            // Calculate matches.
            vector<Match> matches = getMatches(featuresNext, featuresCurr);

            // Find image rotation and ideal homography.
            Mat H, G;
            double rot = findBestRotationCached(i, i+1, prevRot,
                                                matches, images[i+1].size(), H);
            
            // Rotate images and combine.
            Mat imNext = cylinderize(rotateImage(images[i+1], rot));

            Point2f offset;
            combined = combine(H, imNext, combined, G, offset);
            waitKey(1);
            imshow("combined", maskedTo8U(combined));

            // Distort the next image's extracted features.
            Point2f center(images[i - 1].cols / 2.0, images[i+1].rows / 2.0);
            for (size_t f = 0; f < featuresNext.size(); ++f)
            {   
                featuresNext[f].loc = rotatePoint(featuresNext[f].loc,
                                                  center, rot);
                featuresNext[f].loc = ToCylinder(featuresNext[f].loc,
                                                 images[i+1].size(),
                                                 images[i+1].size());
                featuresNext[f].loc = applyHomography(G, featuresNext[f].loc);
            }
            
            // Shift things along in the series.
            featuresCurr = featuresNext;
            prevRot = rot;
    }
    
    combined = maskedTo8U(combined);
    return tightCrop(combined);
}